import os
import logging
import xedocs
from rframe import BaseSchema
from pydantic import BaseSettings
from xedocs.database_interface import DatabaseInterface


logger = logging.getLogger(__name__)
XEDOCS_API_ENV = os.getenv(
    "XEDOCS_API_ENV", os.path.join(xedocs.settings.CONFIG_DIR, "api.env")
)


class ApiSettings(BaseSettings):
    class Config:
        env_prefix = "XEDOCS_API_"
        env_file = XEDOCS_API_ENV

    PRIORITY: int = -1
    URL_TEMPLATE: str = "{base_url}/{version}/{database}/{name}"
    BASE_URL: str = "https://api.xedocs.yossisprojects.com"
    VERSION: str = "v1"
    AUDIENCE: str = "https://api.cmt.xenonnt.org"
    READONLY: bool = False
    TOKEN: str = None
    USERNAME: str = None
    PASSWORD: str = None

    @property
    def token(self):
        if hasattr(self.TOKEN, "access_token"):
            if self.TOKEN.expired:
                self.TOKEN.refresh()
            return self.TOKEN.access_token
        return self.TOKEN

    def login(self):
        import xeauth

        if self.READONLY:
            scopes = ["read:all"]
        else:
            scopes = ["read:all", "write:all"]

        token = xeauth.login(
            username=self.USERNAME,
            password=self.PASSWORD,
            scopes=scopes,
            audience=self.AUDIENCE,
        )

        self.TOKEN = token

    def api_url_for_schema(self, schema: str, database="development"):

        if hasattr(schema, "_ALIAS"):
            schema = schema._ALIAS

        if isinstance(schema, BaseSchema):
            schema = type(schema)

        if not isinstance(schema, str):
            schema = schema.__name__.lower()

        return self.URL_TEMPLATE.format(
            base_url=self.BASE_URL.strip("/"),
            version=self.VERSION.strip("/"),
            name=schema.strip("/"),
            database=database.strip("/"),
        )


class APIDatabase(DatabaseInterface):
    settings: ApiSettings = ApiSettings()
    database: str
    alias: str

    def __init__(
        self, database: str = None, alias: str = None, settings: ApiSettings = None
    ):
        self.database = database
        if alias is None:
            alias = database
        self.alias = alias
        if settings is not None:
            self.settings = settings

    def datasource_for_schema(self, schema):
        import xedocs

        url = self.settings.api_url_for_schema(schema, database=self.database)

        return xedocs.api.api_client(
            url, token=self.settings.TOKEN, authenticator=self.settings
        )



import os
from pathlib import Path
import appdirs
import logging
import pandas as pd

from rframe import BaseSchema


logger = logging.getLogger(__name__)


def xent_collection(**kwargs):
    pass


try:
    import utilix

    uconfig = utilix.uconfig
    from utilix import xent_collection

except ImportError:
    uconfig = None

from pydantic import BaseSettings

from .clock import SimpleClock

dirs = appdirs.AppDirs("xedocs")

XEDOCS_ENV = os.getenv("XEDOCS_ENV", os.path.join(dirs.user_config_dir, "settings.env"))


class Settings(BaseSettings):
    class Config:
        env_prefix = "XEDOCS_"
        env_file = XEDOCS_ENV
        env_file_encoding = "utf-8"

    _DATABASE_INTERFACE_CLASSES = {}

    DATABASES = ["development_db", "straxen_db"]
    DEFAULT_DATABASE = "straxen_db"
    CONFIG_DIR = dirs.user_config_dir
    DATA_DIR = dirs.user_data_dir

    clock = SimpleClock()

    _database_interfaces = {}

    def database_interfaces(self, database: str):
        interfaces = self._database_interfaces.get(database, {})
        interfaces = dict(
            sorted(interfaces.items(), key=lambda x: x[1].settings.PRIORITY)
        )

        return interfaces

    def iter_sources_for_schema(
        self, schema: BaseSchema, interfaces=None, databases=None
    ):
        if databases is None:
            databases = self.DATABASES
        for dbname in databases:
            interfaces = self.database_interfaces(dbname)
            for iname, interface in interfaces.items():
                if interfaces is not None and iname not in interfaces:
                    continue
                source = interface.datasource_for_schema(schema)
                if source is None:
                    continue
                yield (dbname, iname, source)

    def register_databases(self, schema: BaseSchema, databases=None, interfaces=None):
        for dbname, iname, source in self.iter_sources_for_schema(
            schema, interfaces=interfaces, databases=databases
        ):
            for name in (dbname, f"{dbname}_{iname}"):
                if hasattr(schema, name):
                    continue
                schema.register_datasource(source, name=name)

    def default_datasource_for_schema(self, schema, databases=None, interfaces=None):
        if databases is None:
            databases = [self.DEFAULT_DATABASE]
        for _, _, source in self.iter_sources_for_schema(
            schema, databases=databases, interfaces=interfaces
        ):
            return source

    def register_database_interface_class(self, name: str, interface_class):
        self._DATABASE_INTERFACE_CLASSES[name] = interface_class

        if not hasattr(interface_class, "settings"):
            return

        if interface_class.settings.PRIORITY == -1:
            return

        for database in self.DATABASES:
            if database not in self._database_interfaces:
                self._database_interfaces[database] = {}
            try:
                interface = interface_class(database=database)
                self._database_interfaces[database][name] = interface
            except Exception as e:
                logger.debug(
                    f"Could not register {database} interface for {name}. \n Error: {e}"
                )

    def run_doc(self, run_id, fields=("start", "end")):
        if uconfig is None:
            raise KeyError(f"Rundb not configured.")

        rundb = xent_collection()

        if isinstance(run_id, str):
            run_id = int(run_id)

        query = {"number": run_id}

        doc = rundb.find_one(query, projection={f: 1 for f in fields})
        if not doc:
            raise KeyError(f"Run {run_id} not found.")

        return doc

    def run_id_to_time(self, run_id):
        doc = self.run_doc(run_id)
        # use center time of run
        time = doc["start"] + (doc["end"] - doc["start"]) / 2
        return self.clock.normalize_tz(time)

    def run_id_to_interval(self, run_id):
        doc = self.run_doc(run_id)
        start = self.clock.normalize_tz(doc["start"])
        end = self.clock.normalize_tz(doc["end"])
        return start, end

    def extract_time(self, kwargs):
        if "time" in kwargs:
            time = kwargs.pop("time")
        elif "run_id" in kwargs:
            time = self.run_id_to_time(kwargs.pop("run_id"))
        else:
            return None
        time = pd.to_datetime(time)
        time = self.clock.normalize_tz(time)
        return time


settings = Settings()



"""Main module."""

from pathlib import Path
import pandas as pd

from collections import defaultdict
from typing import Dict, List, Type, Union
from rframe import DataAccessor
from tqdm.auto import tqdm

from ._settings import settings
from .schemas import XeDoc

__all__ = [
    "help",
    "get_accessor",
    "find_docs",
    "find_iter",
    "find_df",
    "find_one",
    "insert_docs",
    "list_schemas",
    "all_schemas",
    "schemas_by_category",
    "find_schema",
    "download_db",
]


def find_docs(schema, datasource=None, **labels):
    """find documents by labels

    Args:
        schema (Union[XeDoc,str]): A Xedocs schema or name/alias of one.
        datasource (optional): compatible datasource or name of known source. Defaults to None.
        **labels: label selections
    Returns:
        List[XeDoc]: a list of documents matching selection.
    """

    accessor = get_accessor(schema, datasource)

    return accessor.find_docs(**labels)


def find_iter(schema, datasource=None, **labels):
    """find documents by labels, return iterator

    Args:
        schema (Union[XeDoc,str]): A Xedocs schema or name/alias of one.
        datasource (optional): compatible datasource or name of known source. Defaults to None.
        **labels: label selections
    Returns:
        Iterator[XeDoc]: an iterator over documents matching selection.
    """

    accessor = get_accessor(schema, datasource)

    return accessor.find_iter(**labels)


def find_df(schema, datasource=None, **labels):
    """find dataframe by labels

    Args:
        schema (Union[XeDoc,str]): A Xedocs schema or name/alias of one.
        datasource (optional): compatible datasource or name of known source. Defaults to None.
        **labels: label selections
    Returns:
        DataFrame: a pandas dataframe matching selection.
    """

    accessor = get_accessor(schema, datasource)

    return accessor.find_df(**labels)


def find_one(schema, datasource=None, **labels):
    """find dataframe by labels

    Args:
        schema (Union[XeDoc,str]): A Xedocs schema or name/alias of one.
        datasource (optional): compatible datasource or name of known source. Defaults to None.
        **labels: label selections
    Returns:
        Optional[XeDoc]: The first document matching selection or None.
    """

    accessor = get_accessor(schema, datasource)

    return accessor.find_one(**labels)


def insert_docs(schema: str, docs: Union[list, dict, pd.DataFrame], datasource=None):
    if isinstance(docs, pd.DataFrame):
        docs = docs.reset_index().to_dict(orient="records")
    if not isinstance(docs, list):
        docs = [docs]
    accessor = get_accessor(schema, datasource)

    return accessor.insert(docs)


def list_schemas() -> List[str]:
    return list(XeDoc._XEDOCS)


def all_schemas() -> Dict[str, Type[XeDoc]]:
    return dict(XeDoc._XEDOCS)


def schemas_by_category() -> Dict[str, Dict[str, Type[XeDoc]]]:
    d = defaultdict(dict)
    for name, schema in all_schemas().items():
        d[schema._CATEGORY][name] = schema
    return d


def find_schema(name) -> Type[XeDoc]:

    if isinstance(name, type) and issubclass(name, XeDoc):
        return name

    if not isinstance(name, str):
        raise TypeError(
            f"Schema name must be a string or XeDoc class, not {type(name)}"
        )

    schema = XeDoc._XEDOCS.get(name, None)

    if schema is not None:
        return schema

    for schema in XeDoc._XEDOCS.values():
        if schema.__name__ == name:
            return schema

    raise KeyError(f"Correction with name {name} not found.")


def get_accessor(schema, db=settings.DEFAULT_DATABASE):
    schema = find_schema(schema)

    if not issubclass(schema, XeDoc):
        raise TypeError(
            "Schema must be a subclass of XeDoc" "or the name of a known schema."
        )
    if db is None:
        db = settings.DEFAULT_DATABASE
    if isinstance(db, str):
        accessor = getattr(schema, db)
    else:
        accessor = DataAccessor(schema, db)

    return accessor


def help(schema):
    if isinstance(schema, str):
        schema = find_schema(schema)
    return schema.help()


def default_datasource_for(schema):
    schema = find_schema(schema)

    return settings.default_datasource_for_schema(schema)


def get_api_client(schema, database=settings.DEFAULT_DATABASE):
    interface_class = settings._DATABASE_INTERFACE_CLASSES.get("api", None)
    if interface_class is None:
        raise ValueError("No API interface class found.")
    interface = interface_class(database)
    schema = find_schema(schema)
    return interface.datasource_for_schema(schema)


def download_db(
    dbname=settings.DEFAULT_DATABASE, schemas=None, path=None, batch_size=10_000, verbose=True
):
    """Download data from a remote database to a local database."""
    import tinydb

    if schemas is None:
        schemas = list_schemas()

    if not isinstance(schemas, list):
        schemas = [schemas]

    with tqdm(total=len(schemas)) as pbar:

        for schema in schemas:
            schema = find_schema(schema)
            accessor = get_accessor(schema, dbname)
            pbar.set_description(f"Downloading {schema._ALIAS}")

            basepath = None

            if path is None:
                path = settings.DATA_DIR

            interface = settings._database_interfaces.get(dbname, {}).get(
                "local_repo", None
            )

            if interface is None:
                basepath = Path(path) / dbname / schema._CATEGORY / schema._ALIAS
            else:
                basepath = interface.base_path_for_schema(schema)

            def write_docs(docs, num=1):
                fpath = basepath / f"{num}.json"
                db = tinydb.TinyDB(fpath, create_dirs=True, indent=4)
                table = db.table(schema._ALIAS)
                table.truncate()
                table.insert_multiple(docs)

            filenum = 1
            docs = []
            with tqdm(
                total=accessor.count(), desc=schema._ALIAS, leave=verbose
            ) as pbar2:
                for doc in accessor.find_iter():
                    docs.append(doc.jsonable())
                    pbar2.update(1)
                    if len(docs) >= batch_size:
                        write_docs(docs, filenum)
                        docs = []
                        filenum += 1
                if len(docs) > 0:
                    write_docs(docs, filenum)

            pbar.update(1)



from abc import ABC, abstractmethod
from typing import Iterable
from rframe import DataAccessor
from collections import UserDict

from .schemas import XeDoc
from ._settings import settings
from .utils import DatasetCollection


class DatabaseInterface(ABC):
    @abstractmethod
    def datasource_for_schema(schema: XeDoc):
        pass

    def __getitem__(self, key):
        schema = XeDoc._XEDOCS.get(key, None)
        if schema is None:
            raise KeyError(f"No schema named {key} found")

        source = self.datasource_for_schema(schema)

        if source is None:
            raise KeyError(f"No datasource for {key} found")

        return DataAccessor(schema, source)

    def __getattr__(self, attr):
        if attr in XeDoc._XEDOCS.keys():
            return self[attr]
        raise AttributeError(attr)

    def __dir__(self):
        return list(XeDoc._XEDOCS.keys()) + super().__dir__()


class DatabasesCollection(UserDict):
    def __getitem__(self, key):
        if key in settings.DATABASES:
            interfaces = settings.database_interfaces(key)
            return DatasetCollection(interfaces)
        raise KeyError(f"No database named {key} found.")

    def __getattr__(self, attr):
        if attr in settings.DATABASES:
            return self[attr]
        raise AttributeError(attr)

    def __dir__(self) -> Iterable[str]:
        return list(settings.DATABASES)


def load_db_interfaces():

    from .entrypoints import load_entry_points

    hooks = load_entry_points("xedocs_db_interfaces")

    for name, hook in hooks.items():
        if isinstance(hook, type) and issubclass(hook, DatabaseInterface):
            settings.register_database_interface_class(name, hook)



import os
import xedocs
import pymongo
import logging

from pydantic import BaseSettings
from xedocs.schemas import XeDoc
from xedocs.database_interface import DatabaseInterface


XEDOCS_MONGO_ENV = os.getenv(
    "XEDOCS_MONGO_ENV", os.path.join(xedocs.settings.CONFIG_DIR, "mongo.env")
)
logger = logging.getLogger(__name__)


class MongoSettings(BaseSettings):
    class Config:
        env_prefix = "XEDOCS_MONGO_"
        env_file = XEDOCS_MONGO_ENV

    PRIORITY: int = -1
    URL: str = "mongodb://localhost:27017"

    def client(self):
        return pymongo.MongoClient(self.URL)


class MongoDatabase(DatabaseInterface):
    settings: MongoSettings = MongoSettings()
    database: str
    alias: str
    _client: pymongo.MongoClient = None

    def __init__(
        self,
        database: str = None,
        alias: str = None,
        settings: MongoSettings = None,
        client: pymongo.MongoClient = None,
    ):
        self.database = database
        if alias is None:
            alias = database
        self.alias = alias
        if settings is not None:
            self.settings = settings
        if client is not None:
            self._client = client

    @property
    def client(self):
        if self._client is None:
            self._client = self.settings.client()
        return self._client

    def datasource_for_schema(self, schema: XeDoc):
        return self.client[self.database][schema._ALIAS]



import os
import xedocs
import logging
from utilix import xent_collection, uconfig

from xedocs.database_interface import DatabaseInterface
from pydantic import BaseSettings


XEDOCS_UTILIX_ENV = os.getenv(
    "XEDOCS_UTILIX_ENV", os.path.join(xedocs.settings.CONFIG_DIR, "utilix.env")
)
logger = logging.getLogger(__name__)


class UtilixSettings(BaseSettings):
    class Config:
        env_prefix = "XEDOCS_UTILIX_"
        env_file = XEDOCS_UTILIX_ENV

    PRIORITY: int = 1
    DATABASE_NAMES = {"straxen_db": "xedocs", "development_db": "xedocs-dev"}


class UtilixDatabase(DatabaseInterface):
    settings: UtilixSettings = UtilixSettings()
    database: str
    alias: str

    def __init__(
        self, database: str = None, alias: str = None, settings: UtilixSettings = None
    ):
        if alias is None:
            alias = database
        self.alias = alias
        if settings is not None:
            self.settings = settings
        database = self.settings.DATABASE_NAMES.get(database, database)
        self.database = database

    def datasource_for_schema(self, schema):
        if uconfig is None:
            return None
        collection_name = schema._ALIAS
        return xent_collection(collection=collection_name, database=self.database)
